
#include "mapinc.h"

static const uint32 WRAMSIZE = 8*1024;
static const uint32 CHRRAMSIZE = 32*1024;

static uint8 *WRAM = NULL;
static uint8 *CHRRAM = NULL;

static uint8 MMC3_CMD;
static uint8 MMC3_BANK[8];
static uint8 MMC3_MIR;
static uint8 MMC3_WRAM;
static uint8 IRQ_RLD;
static uint8 IRQ_CNT;
static uint8 IRQ_LAT;
static uint8 IRQ_ENA;

static SFORMAT QMMC3W194_StateRegs[] =
{
	{ MMC3_BANK, 8, "REGS" },
	{ &MMC3_CMD, 1, "CMD" },
	{ &MMC3_MIR, 1, "MIR" },
	{ &MMC3_WRAM, 1, "WRAM" },
	{ &IRQ_RLD, 1, "IRQR" },
	{ &IRQ_CNT, 1, "IRQC" },
	{ &IRQ_LAT, 1, "IRQL" },
	{ &IRQ_ENA, 1, "IRQE" },
	{ 0 }
};

static void QMMC3W194_SetChrBank(uint32 A, uint8 V) {
	switch ((MMC3_MIR & 0xc0) >> 6) {
	case 0:
		setchr1r(0, A, V);
		break;
	case 1:
		if (V < 2) {
			setchr1r(0x10, A, V);
		} else {
			setchr1r(0, A, V);
		}
		break;
	case 2:
		if (V < 8) {
			setchr1r(0x10, A, V);
		} else {
			setchr1r(0, A, V);
		}
		break;
	case 3:
		if (V < 32) {
			setchr1r(0x10, A, V);
		} else {
			setchr1r(0, A, V);
		}
		break;
	}
}

static void QMMC3W194_ResetPrgBank(uint8 V) {
	if (V & 0x40) {
		setprg8(0xc000, MMC3_BANK[6] & 0x3f);
		setprg8(0x8000, 0x3e);
	} else {
		setprg8(0x8000, MMC3_BANK[6] & 0x3f);
		setprg8(0xc000, 0x3e);
	}
	setprg8(0xa000, MMC3_BANK[7] & 0x3f);
	setprg8(0xe000, 0x3f);
}

static void QMMC3W194_ResetChrBank(uint8 V) {
	if (V & 0x80) {
		QMMC3W194_SetChrBank(0x1000, (MMC3_BANK[0] & 0xfe) | 0x00);
		QMMC3W194_SetChrBank(0x1400, (MMC3_BANK[0] & 0xfe) | 0x01);
		QMMC3W194_SetChrBank(0x1800, (MMC3_BANK[1] & 0xfe) | 0x00);
		QMMC3W194_SetChrBank(0x1c00, (MMC3_BANK[1] & 0xfe) | 0x01);
		QMMC3W194_SetChrBank(0x0000, MMC3_BANK[2] & 0xff);
		QMMC3W194_SetChrBank(0x0400, MMC3_BANK[3] & 0xff);
		QMMC3W194_SetChrBank(0x0800, MMC3_BANK[4] & 0xff);
		QMMC3W194_SetChrBank(0x0c00, MMC3_BANK[5] & 0xff);
	} else {
		QMMC3W194_SetChrBank(0x0000, (MMC3_BANK[0] & 0xfe) | 0x00);
		QMMC3W194_SetChrBank(0x0400, (MMC3_BANK[0] & 0xfe) | 0x01);
		QMMC3W194_SetChrBank(0x0800, (MMC3_BANK[1] & 0xfe) | 0x00);
		QMMC3W194_SetChrBank(0x0c00, (MMC3_BANK[1] & 0xfe) | 0x01);
		QMMC3W194_SetChrBank(0x1000, MMC3_BANK[2] & 0xff);
		QMMC3W194_SetChrBank(0x1400, MMC3_BANK[3] & 0xff);
		QMMC3W194_SetChrBank(0x1800, MMC3_BANK[4] & 0xff);
		QMMC3W194_SetChrBank(0x1c00, MMC3_BANK[5] & 0xff);
	}
	setmirror((MMC3_MIR & 1) ? MI_H : MI_V);
}

DECLFW(QMMC3W194_MapperWrite) {
	switch (A & 0xe001) {
	case 0x8000:
		if ((V & 0x40) != (MMC3_CMD & 0x40)) {
			QMMC3W194_ResetPrgBank(V);
		}
		if ((V & 0x80) != (MMC3_CMD & 0x80)) {
			QMMC3W194_ResetChrBank(V);
		}
		MMC3_CMD = V;
		break;
	case 0x8001:
		MMC3_BANK[MMC3_CMD & 0x7] = V;
		switch (MMC3_CMD & 0x07) {
		case 0:
			if (MMC3_CMD & 0x80) {
				QMMC3W194_SetChrBank(0x1000, (V & 0xfe) | 0x00);
				QMMC3W194_SetChrBank(0x1400, (V & 0xfe) | 0x01);
			} else {
				QMMC3W194_SetChrBank(0x0000, (V & 0xfe) | 0x00);
				QMMC3W194_SetChrBank(0x0400, (V & 0xfe) | 0x01);
			}
			break;
		case 1:
			if (MMC3_CMD & 0x80) {
				QMMC3W194_SetChrBank(0x1800, (V & 0xfe) | 0x00);
				QMMC3W194_SetChrBank(0x1c00, (V & 0xfe) | 0x01);
			} else {
				QMMC3W194_SetChrBank(0x0800, (V & 0xfe) | 0x00);
				QMMC3W194_SetChrBank(0x0c00, (V & 0xfe) | 0x01);
			}
			break;
		case 2:
			if (MMC3_CMD & 0x80) {
				QMMC3W194_SetChrBank(0x0000, V & 0xff);
			} else {
				QMMC3W194_SetChrBank(0x1000, V & 0xff);
			}
			break;
		case 3:
			if (MMC3_CMD & 0x80) {
				QMMC3W194_SetChrBank(0x0400, V & 0xff);
			} else {
				QMMC3W194_SetChrBank(0x1400, V & 0xff);
			}
			break;
		case 4:
			if (MMC3_CMD & 0x80) {
				QMMC3W194_SetChrBank(0x0800, V & 0xff);
			} else {
				QMMC3W194_SetChrBank(0x1800, V & 0xff);
			}
			break;
		case 5:
			if (MMC3_CMD & 0x80) {
				QMMC3W194_SetChrBank(0x0c00, V & 0xff);
			} else {
				QMMC3W194_SetChrBank(0x1c00, V & 0xff);
			}
			break;
		case 6:
			if (MMC3_CMD & 0x40) {
				setprg8(0xc000, V & 0x3f);
			} else {
				setprg8(0x8000, V & 0x3f);
			}
			break;
		case 7:
			setprg8(0xa000, V & 0x3f);
			break;
		}
		break;
	case 0xa000:
		MMC3_MIR = V;
		QMMC3W194_ResetChrBank(MMC3_CMD);
		setmirror((V & 1) ? MI_H : MI_V);
		break;
	case 0xa001:
		MMC3_WRAM = V;
		break;
	case 0xc000:
		IRQ_LAT = V;
		break;
	case 0xc001:
		IRQ_RLD = 1;
		break;
	case 0xe000:
		X6502_IRQEnd(FCEU_IQEXT);
		IRQ_ENA = 0;
		break;
	case 0xe001:
		IRQ_ENA = 1;
		break;
	}
}

DECLFR(QMMC3W194_WramRead) {
	if ((MMC3_WRAM & 0x80) == 0x00) {
		return 0x00;
	}
	return CartBR(A);
}

DECLFW(QMMC3W194_WramWrite) {
	if ((MMC3_WRAM & 0xc0) == 0x80) {
		CartBW(A,V);
	}
}

void QMMC3W194_HBlank(void) {
	int count = IRQ_CNT;
	if (!count || IRQ_RLD) {
		IRQ_CNT = IRQ_LAT;
		IRQ_RLD = 0;
	} else {
		IRQ_CNT--;
	}
	if (count && !IRQ_CNT) {
		if (IRQ_ENA) {
			X6502_IRQBegin(FCEU_IQEXT);
		}
	}
}

void QMMC3W194_Reset(void) {
	;
}

void QMMC3W194_Power(void) {

	// Register a handler for memory access.
	SetWriteHandler(0x8000, 0xffff, QMMC3W194_MapperWrite);
	SetReadHandler(0x8000, 0xffff, CartBR);
	SetWriteHandler(0x6000, 0x7fff, QMMC3W194_WramWrite);
	SetReadHandler(0x6000, 0x7fff, QMMC3W194_WramRead);

	FCEU_CheatAddRAM(WRAMSIZE>>10, 0x6000, WRAM);

	// initialize a inner registers.
	MMC3_CMD = 0;
	IRQ_CNT = 0;
	IRQ_LAT = 0;
	IRQ_ENA = 0;

	MMC3_BANK[0] = 0;
	MMC3_BANK[1] = 2;
	MMC3_BANK[2] = 4;
	MMC3_BANK[3] = 5;
	MMC3_BANK[4] = 6;
	MMC3_BANK[5] = 7;
	MMC3_BANK[6] = 0;
	MMC3_BANK[7] = 1;

	// Bank settings for fixed area.
	setprg8r(0x10, 0x6000, 0);
	setprg8(0x8000, MMC3_BANK[6] & 0x3f);
	setprg8(0xa000, MMC3_BANK[7] & 0x3f);
	setprg8(0xc000, 0x3e);
	setprg8(0xe000, 0x3f);

	setchr1(0x0000, (MMC3_BANK[0] & 0xfe) | 0x00);
	setchr1(0x0400, (MMC3_BANK[0] & 0xfe) | 0x01);
	setchr1(0x0800, (MMC3_BANK[1] & 0xfe) | 0x00);
	setchr1(0x0c00, (MMC3_BANK[1] & 0xfe) | 0x01);
	setchr1(0x1000, MMC3_BANK[2] & 0xff);
	setchr1(0x1400, MMC3_BANK[3] & 0xff);
	setchr1(0x1800, MMC3_BANK[4] & 0xff);
	setchr1(0x1c00, MMC3_BANK[5] & 0xff);

	// Nametable milloring settings at startup.
	setmirror(MI_V);

	QMMC3W194_Reset();
}

void QMMC3W194_Close(void) {
	// Free extended memory.
	if (WRAM) FCEU_gfree(WRAM);
	WRAM = NULL;
	if (CHRRAM) FCEU_gfree(CHRRAM);
	CHRRAM = NULL;
}

void QMMC3W194_Init(CartInfo *info) {

	// Allocate extended memory.
	WRAM = (uint8*)FCEU_gmalloc(WRAMSIZE);
	SetupCartPRGMapping(0x10, WRAM, WRAMSIZE, 1);
	AddExState(WRAM, WRAMSIZE, 0, "WRAM");

	CHRRAM = (uint8*)FCEU_gmalloc(CHRRAMSIZE);
	SetupCartCHRMapping(0x10, CHRRAM, CHRRAMSIZE, 1);
	AddExState(CHRRAM, CHRRAMSIZE, 0, "CHRR");

	// Register memory for save.
	info->SaveGame[0] = WRAM;
	info->SaveGameLen[0] = WRAMSIZE;

	// Register internal register state
	AddExState(QMMC3W194_StateRegs, 0, 0, 0);

	// Register callback.
	info->Power = QMMC3W194_Power;
	info->Reset = QMMC3W194_Reset;
	info->Close = QMMC3W194_Close;

	// register HBlank hook.
	GameHBIRQHook = QMMC3W194_HBlank;
}

