import argparse
import struct
import os
import string

CH_MAX = 4
EP_MAX = 15
TONE_BITS = 4
LEN_BITS = 8 - TONE_BITS

BASE_TICK = 192
DEF_TEMPO = 120
OCTAVE_MIN = 2
OCTAVE_MAX = 7
OCTAVE_WIDTH = 12

TYPE_UNKNOWN   =  0
TYPE_NOTE      =  1
TYPE_REST      =  2
TYPE_EXTEND    =  3
TYPE_REG1      =  4
TYPE_REG2      =  5
TYPE_TONE_BASE =  6
TYPE_LEN_TABLE =  7
TYPE_ENV_BASE  =  8
TYPE_TEMPO     =  9
TYPE_PERCUSS   = 10
TYPE_VOLUME    = 11
TYPE_REPEAT    = 12
TYPE_REPEAT_W  = 13
TYPE_JMP       = 14
TYPE_CALL      = 15
TYPE_RET       = 16
TYPE_LOOP_CNT  = 17
TYPE_END       = 18
TYPE_LABEL     = 32
TYPE_SKIP      = 33
TYPE_SKIP_W    = 34

#           value        len        quant
# NOTE      音程         音長       quant値
# REST      0            音長       quant値
# EXTEND    0            音長       quant値
# REG1      設定値       0          0
# REG2      設定値       0          0
# TONE_BASE 音程(下限)   音程(上限) 無効フラグ
# LEN_TABLE LTable番号   0          無効フラグ
# ENV_BASE  env番号      0          0
# TEMPO     テンポ値     0          0
# PERCUSS   打音番号     0          0
# VOLUME    変更量       0          0
# REPEAT    繰り返し回数 ラベル番号 short/long
# REPEAT_W  繰り返し回数 ラベル番号 short/long
# JMP       参照番号     ラベル番号 short/long
# CALL      参照番号     ラベル番号 short/long
# RET       0            0          0
# LOOP_CNT  0            0          0
# END       0            0          0
# LABEL     参照番号     ラベル番号 0
# SKIP      0            ラベル番号 0
#
# ラベル番号は トラックループ:0 その他は1以上

CH_CHR = ''
for i in range(CH_MAX):
	CH_CHR += chr(65+i)

TONE_RANGE = (1<<TONE_BITS)-2

TONE_CHR = 'cdefgab'
TONE_INDEX = [ 0, 2, 4, 5, 7, 9, 11 ]

LENTBL_SIZE = (1<<LEN_BITS) - (32>>TONE_BITS)

# Python の List は実質的には可変長配列なので
# 自前で単方向リストを実装する

class event_t:
	def __init__(self, type, value = 0, length = 0, quant = 0):
		self.next = None
		self.type = type
		self.value = value
		self.length = length
		self.quant = quant

class seq_t:
	def __init__(self):
		self.events = []
		self.head = None
		self.tail = None
		self.last_note = None
		self.base_tone = 4*12
		self.def_len = 4
		self.quant = 7
		self.label = 1
		self.save_base_tone = 0
		self.save_base_tone_w = 0
		self.save_def_len = 0
		self.save_def_len_w = 0
		self.save_quant = 0
		self.save_quant_w = 0
		self.exist_skip = False
		self.exist_skip_w = False
		self.assigned_label = 0
		self.assigned_label_w = 0
		self.nest = 0
		self.nest_w = 0
		self.ref_ev = None
		self.max_ref_num = 0

	def appendEvent(self, type, value = 0, length = 0, quant = 0):
		ev = event_t(type, value, length, quant)
		self.events.append(ev)
		if self.tail:
			self.tail.next = ev
		else:
			self.head = ev
		self.tail = ev

	def insertEvent(self, prev, type, value = 0, length = 0, quant = 0):
		ev = event_t(type, value, length, quant)
		self.events.append(ev)
		if self.tail == prev:
			self.tail = ev
		if prev:
			ev.next = prev.next
			prev.next = ev
		else:
			ev.next = self.head
			self.head = ev

class lentbl_t:
	def __init__(self):
		self.length = None
		self.quant = None
		self.count = 0
		self.base = 0
		self.share_count = 0

	def append(self, length, quant):
		if self.count >= LENTBL_SIZE:
			value2 = (length << 8) | quant
			i = len(self.length)
			while i > 0:
				value1 = (self.length[i-1] << 8) | self.quant[i-1]
				if value1 == value2:
					return True
				i -= 1
			return False
		elif self.count > 0:
			value2 = (length << 8) | quant
			i = self.count
			while i > 0:
				value1 = (self.length[i-1] << 8) | self.quant[i-1]
				if value1 == value2:
					return True
				if value1 <= value2:
					break
				i -= 1
			self.length.insert(i, length)
			self.quant.insert(i, quant)
			self.count += 1
		else:
			self.length = [ length ]
			self.quant = [ quant ]
			self.count = 1
		return True

	def commonCount(self, tbl):
		count = 0
		for i in range(self.count):
			for j in range(tbl.count):
				if self.length[i] == tbl.length[j] and self.quant[i] == tbl.quant[j]:
					count += 1
		return count

	def merge(self, tbl):
		for j in range(tbl.count):
			value1 = -1
			value2 = (tbl.length[j] << 8) | tbl.quant[j]
			i = self.count
			while i > 0:
				value1 = (self.length[i-1] << 8) | self.quant[i-1]
				if value1 <= value2:
					break
				i -= 1
			if value1 != value2:
				self.length.insert(i, tbl.length[j])
				self.quant.insert(i, tbl.quant[j])
				self.count += 1

class env_t:
	def __init__(self):
		self.data = []
		self.vol = []
		self.tc = []
		self.dt = []
		self.size = 0
		self.loop = 0
		self.base = 0

class rhythm_t:
	def __init__(self):
		self.name = ''
		self.data = 0
		self.quant = 8

class score_t:
	def __init__(self):
		self.lentbl = []
		self.env = []
		self.rhythm = []
		self.ch = []
		self.tempo_1st = 0
		self.tempo_max = 0

		for i in range(CH_MAX):
			self.ch.append(seq_t())
		for i in range(EP_MAX):
			self.env.append(env_t())

class call_t:
	def __init__(self, ev):
		self.ev = ev
		self.count = 0

class flow_t:
	def __init__(self):
		self.ev = None
		self.low = 0
		self.high = 0
		self.uninit_tone = True
		self.ltbl = None
		self.cont_len = True

class opttbl_t:
	def __init__(self):
		self.near = flow_t()
		self.far = flow_t()
		link1 = None
		link2 = None

def parseArg():
	parser = argparse.ArgumentParser()
	parser.add_argument('mmlfile', help='mml file name')
	parser.add_argument('cmdfile', help='cmd file name')
	parser.add_argument('datfile', help='data file name')
	parser.add_argument('prefix', help='prefix')

	params = parser.parse_args()

	return params

def parseInt(line, pos):
	value = 0
	if line[pos] == '$':
		pos += 1
		while (index := string.hexdigits.find(line[pos])) >= 0:
			if index >= 16:
				index -= 6
			value = value * 16 + index
			pos += 1
	else:
		while (index := string.digits.find(line[pos])) >= 0:
			value = value * 10 + index
			pos += 1
	return pos, value

def parseSignedInt(line, pos):
	minus = False
	if line[pos] == '+':
		pos += 1
	if line[pos] == '-':
		minus = True
		pos += 1
	pos, value = parseInt(line, pos)
	if minus:
		value = -value
	return pos, value

def parseLength(line, pos, def_len):
	length = 0
	is_error = False
	while True:
		c = line[pos]
		if c.isdigit() or c == '$':
			pos, len1 = parseInt(line, pos)
			if BASE_TICK % len1:
				is_error = True
			len1 = BASE_TICK // len1
		else:
			len1 = BASE_TICK // def_len
		len2 = len1
		while line[pos] == '.':
			pos += 1
			if len2 % 2:
				is_error = True
			len2 //= 2
			len1 += len2
		length += len1

		if line[pos] != '^':
			break
		pos += 1
	if is_error:
		return pos, -1

	return pos, length

def skipChar(line, pos, c):
	while line[pos].isspace():
		pos += 1
	if line[pos] == c:
		pos += 1
	while line[pos].isspace():
		pos += 1
	return pos

def parseDefine(score, def_name, line, pos, line_num):
	if def_name == 'EP':
		# エンベロープ定義
		while line[pos].isspace():
			pos += 1
		pos, index = parseInt(line, pos)
		if index >= 1 and index <= EP_MAX:
			env = score.env[index-1]
			pos = skipChar(line, pos, ',')
			env.loop = -1
			c = line[pos]
			while c.isdigit() or c == '$' or c == '+' or c == '-' or c == ':':
				if c == ':':
					pos += 1
					env.loop = len(env.vol)
				else:
					tc = 0
					dt = 0
					pos, value = parseSignedInt(line, pos)
					while line[pos].isspace():
						pos += 1
					if line[pos] == '/':
						pos += 1
						while line[pos].isspace():
							pos += 1
						c = line[pos]
						if c.isdigit() or c == '$':
							pos, tc = parseInt(line, pos)
							while line[pos].isspace():
								pos += 1
							if line[pos] == '/':
								pos += 1
								while line[pos].isspace():
									pos += 1
								c = line[pos]
								if c.isdigit() or c == '$' or c == '+' or c == '-':
									pos, dt = parseSignedInt(line, pos)
					env.vol.append(value)
					env.tc.append(tc)
					env.dt.append(dt)
				pos = skipChar(line, pos, ',')
				c = line[pos]
		else:
			print(f'{line_num}: invalid index of EP.')
	elif def_name == 'RP':
		# リズム定義
		rhythm = rhythm_t()
		score.rhythm.append(rhythm)

		pos = skipChar(line, pos, ',')
		pos2 = pos
		while line[pos].isalpha() or line[pos] == '_':
			pos += 1
		rhythm.name = line[pos2:pos].upper()

		pos = skipChar(line, pos, ',')
		pos, value = parseInt(line, pos)
		rhythm.data |= (value & 7) << 12

		pos = skipChar(line, pos, ',')
		pos, value = parseInt(line, pos)
		rhythm.data |= (value & 0x8f) << 8

		pos = skipChar(line, pos, ',')
		pos, value = parseInt(line, pos)
		rhythm.data |= (value & 15) << 4

		pos = skipChar(line, pos, ',')
		pos, value = parseInt(line, pos)
		rhythm.data |= value & 15

		pos = skipChar(line, pos, ',')
		if line[pos].isdigit():
			pos, value = parseInt(line, pos)
			rhythm.quant = value
		else:
			rhythm.quant = 0
	else:
		# 未定義
		print(f'{line_num}: syntax error. ({def_name})')

def parseScore(score, ch_bits, line, pos, line_num):
	c = line[pos]
	while c != '\0':
		if c.isspace():
			pos += 1
		elif c == ';':
			# コメント
			pos = len(line)-1
		elif c == '@':
			pos += 1
			c = line[pos]
			if c == 'x':
				# レジスタ1指定
				pos += 1
				c = line[pos]
				if c.isdigit() or c == '$':
					pos, value = parseInt(line, pos)
					if value >= 0 and value < 256:
						for i, ch in enumerate(score.ch):
							if ch_bits & (1 << i):
								ch.appendEvent(TYPE_REG1, value, 0, 0)
					else:
						print(f'{line_num}: invalid value in reg1.')
				else:
					print(f'{line_num}: no value in reg1.')
			elif c == 'y':
				# レジスタ2指定
				pos += 1
				c = line[pos]
				if c.isdigit() or c == '$':
					pos, value = parseInt(line, pos)
					if value >= 0 and value < 256:
						for i, ch in enumerate(score.ch):
							if ch_bits & (1 << i):
								ch.appendEvent(TYPE_REG2, value, 0, 0)
					else:
						print(f'{line_num}: invalid value in reg2.')
				else:
					print(f'{line_num}: no value in reg2.')
			elif c == 'v':
				# 音量変更
				pos += 1
				value = 0
				if line[pos] == '+':
					pos += 1
					if line[pos] == '+':
						pos += 1
						value = 2
					else:
						value = 1
				elif line[pos] == '-':
					pos += 1
					if line[pos] == '-':
						pos += 1
						value = -2
					else:
						value = -1
				if value != 0:
					for i, ch in enumerate(score.ch):
						if ch_bits & (1 << i):
							ch.appendEvent(TYPE_VOLUME, value, 0, 0)
				else:
					print(f'{line_num}: invalid value in volume.')
			elif line[pos:pos+2] == 'ev':
				# エンベロープ指定
				pos += 2
				c = line[pos]
				if c.isdigit() or c == '$':
					pos, value = parseInt(line, pos)
					if value >= 0 and value < 256:
						for i, ch in enumerate(score.ch):
							if ch_bits & (1 << i):
								ch.appendEvent(TYPE_ENV_BASE, value, 0, 0)
					else:
						print(f'{line_num}: invalid value in volume envelope.')
				else:
					print(f'{line_num}: no value in volume envelope.')
			elif line[pos:pos+2] == 'lc':
				# ループカウンタ
				pos += 2
				for i, ch in enumerate(score.ch):
					if ch_bits & (1 << i):
						ch.appendEvent(TYPE_LOOP_CNT, 0, 0, 0)
			elif c == '[':
				# 広域リピート(開始)
				pos += 1
				for i, ch in enumerate(score.ch):
					if ch_bits & (1 << i):
						if ch.nest_w < 1:
							ch.assigned_label_w = ch.label
							ch.exist_skip_w = False
							ch.appendEvent(TYPE_LABEL, 0, ch.label, 0)
							ch.label += 1
							ch.nest_w += 1
						else:
							print(f'{line_num}: repeat nest over.')
			elif c == '|':
				# 広域リピート(スキップ)
				pos += 1
				for i, ch in enumerate(score.ch):
					if ch_bits & (1 << i):
						ch.appendEvent(TYPE_SKIP_W, 0, ch.assigned_label_w, 0)
						ch.save_base_tone_w = ch.base_tone
						ch.save_def_len_w = ch.def_len
						ch.save_quant_w = ch.quant
						ch.exist_skip_w = True
			elif c == ']':
				# 広域リピート(終了)
				pos += 1
				c = line[pos]
				if c.isdigit() or c == '$':
					pos, value = parseInt(line, pos)
					if value > 0 and value <= 127:
						for i, ch in enumerate(score.ch):
							if ch_bits & (1 << i):
								ch.nest_w -= 1
								ch.appendEvent(TYPE_REPEAT_W, value, ch.assigned_label_w, 0)
								if ch.exist_skip_w:
									ch.base_tone = ch.save_base_tone_w
									ch.def_len = ch.save_def_len_w
									ch.quant = ch.save_quant_w
					else:
						print(f'{line_num}: invalid value in repeat.')
				else:
					print(f'{line_num}: no value in repeat.')
			else:
				# 未定義
				print(f'{line_num}: syntax error. @{c}')
				pos += 1
		elif (index := TONE_CHR.find(c)) >= 0:
			# 音符
			value = TONE_INDEX[index]
			pos += 1
			if line[pos] == '+':
				value += 1
				pos += 1
			elif line[pos] == '-':
				value -= 1
				pos += 1

			pos2 = pos;
			for i, ch in enumerate(score.ch):
				if ch_bits & (1 << i):
					pos, length = parseLength(line, pos2, ch.def_len)
					if length > 0:
						if length > BASE_TICK:
							length2 = length
							while length2 > BASE_TICK:
								length2 -= BASE_TICK
							ch.appendEvent(TYPE_NOTE, value + ch.base_tone, length2, 8)
							length2 += BASE_TICK
							while length2 < length:
								ch.appendEvent(TYPE_EXTEND, 0, BASE_TICK, 8)
								length2 += BASE_TICK
							ch.appendEvent(TYPE_EXTEND, 0, BASE_TICK, ch.quant)
						else:
							ch.appendEvent(TYPE_NOTE, value + ch.base_tone, length, ch.quant)
						ch.last_note = ch.tail
					else:
						print(f'{line_num}: invalid tone length ({line[pos2:pos]})')
			if score.tempo_1st == 0:
				score.tempo_1st = DEF_TEMPO
			if score.tempo_max == 0:
				score.tempo_max = DEF_TEMPO
		elif c == 'r':
			# 休符
			pos += 1

			pos2 = pos
			for i, ch in enumerate(score.ch):
				if ch_bits & (1 << i):
					pos, length = parseLength(line, pos2, ch.def_len)
					if length > 0:
						while length > BASE_TICK:
							ch.appendEvent(TYPE_REST, 0, BASE_TICK, 8)
							length -= BASE_TICK
						ch.appendEvent(TYPE_REST, 0, length, ch.quant)
						ch.last_note = ch.tail
					else:
						print(f'{line_num}: invalid tone length ({line[pos2:pos]})')
			if score.tempo_1st == 0:
				score.tempo_1st = DEF_TEMPO
			if score.tempo_max == 0:
				score.tempo_max = DEF_TEMPO
		elif c == '!':
			# リズム
			pos += 1

			name = ''
			c = line[pos]
			while c.isalpha() or c == '_':
				name += c.upper()
				pos += 1
				c = line[pos]

			pos2 = pos
			i = next((i for i, e in enumerate(score.rhythm) if e.name == name), -1)
			if i >= 0:
				if ch_bits & (1 << 3): # 2a03ノイズ(Dチャンネル)のみ
					rhythm = score.rhythm[i]
					ch = score.ch[3]
					pos, length = parseLength(line, pos2, ch.def_len)
					if length > 0:
						quant = rhythm.quant if rhythm.quant > 0 else ch.quant
						ch.appendEvent(TYPE_NOTE, i+32+2*12, length, quant)
					else:
						print(f'{line_num}: invalid tone length ({line[pos2:pos]})')
			if score.tempo_1st == 0:
				score.tempo_1st = DEF_TEMPO
			if score.tempo_max == 0:
				score.tempo_max = DEF_TEMPO
		elif c == '&':
			# タイ
			pos += 1

			pos2 = pos
			for i, ch in enumerate(score.ch):
				if ch_bits & (1 << i):
					pos, length = parseLength(line, pos2, ch.def_len)
					if length > 0:
						if ch.last_note:
							ch.last_note.quant = 8
							if length > BASE_TICK:
								length2 = length
								while length2 > BASE_TICK:
									length2 -= BASE_TICK
								ch.appendEvent(TYPE_EXTEND, 0, length2, 8)
								length2 += BASE_TICK
								while length2 < length:
									ch.appendEvent(TYPE_EXTEND, 0, BASE_TICK, 8)
									length += BASE_TICK
								ch.appendEvent(TYPE_EXTEND, 0, BASE_TICK, ch.quant)
							else:
								ch.appendEvent(TYPE_EXTEND, 0, length, ch.quant)
							ch.last_note = ch.tail
					else:
						print(f'{line_num}: invalid tone length ({line[pos2:pos]})')
		elif c == 'p':
			# 打音
			pos += 1

			value = 0
			c = line[pos]
			if c.isdigit() or c == '$':
				pos, value = parseInt(line, pos)

			for i, ch in enumerate(score.ch):
				if ch_bits & (1 << i):
					ch.appendEvent(TYPE_PERCUSS, value, 0, 0)
		elif c == '>':
			# オクターブ上昇
			pos += 1

			for i, ch in enumerate(score.ch):
				if ch_bits & (1 << i):
					ch.base_tone += OCTAVE_WIDTH
		elif c == '<':
			# オクターブ下降
			pos += 1

			for i, ch in enumerate(score.ch):
				if ch_bits & (1 << i):
					ch.base_tone -= OCTAVE_WIDTH
		elif c == 'o':
			# オクターブ指定
			pos += 1

			c = line[pos]
			if c.isdigit() or c == '$':
				pos, value = parseInt(line, pos)
				if value >= OCTAVE_MIN and value <= OCTAVE_MAX:
					for i, ch in enumerate(score.ch):
						if ch_bits & (1 << i):
							ch.base_tone = value * OCTAVE_WIDTH
				else:
					print(f'{line_num}: invalid value in octave.')
			else:
				print(f'{line_num}: no value in octave.')
		elif c == 't':
			# テンポ指定
			pos += 1

			c = line[pos]
			if c.isdigit() or c == '$':
				pos, value = parseInt(line, pos)
				n = 1 if score.tempo_1st == 0 else 0
				for i, ch in enumerate(score.ch):
					if ch_bits & (1 << i):
						ch.appendEvent(TYPE_TEMPO, value, 0, n)
				if score.tempo_1st == 0:
					score.tempo_1st = value
				if value > score.tempo_max:
					score.tempo_max = value
			else:
				print(f'{line_num}: no value in tempo.')
		elif c == 'l':
			# 省略時の音長指定
			pos += 1

			c = line[pos]
			if c.isdigit() or c == '$':
				pos, value = parseInt(line, pos)
				if (BASE_TICK % value) == 0:
					for i, ch in enumerate(score.ch):
						if ch_bits & (1 << i):
							ch.def_len = value
				else:
					print(f'{line_num}: invalid value in default length.')
			else:
				print(f'{line_num}: no value in default length.')
		elif c == 'q':
			# クォンタイズ指定
			pos += 1

			c = line[pos]
			if c.isdigit() or c == '$':
				pos, value = parseInt(line, pos)
				if value > 0 and value <= 8:
					for i, ch in enumerate(score.ch):
						if ch_bits & (1 << i):
							ch.quant = value
				else:
					print(f'{line_num}: invalid value in quantize.')
			else:
				print(f'{line_num}: no value in quantize.')
		elif c == 'm':
			# モチーフ呼び出し
			pos += 1

			c = line[pos]
			if c.isdigit() or c == '$':
				pos, value = parseInt(line, pos)
				if value > 0:
					for i, ch in enumerate(score.ch):
						if ch_bits & (1 << i):
							ch.appendEvent(TYPE_CALL, value, 0, 0)
				else:
					print(f'{line_num}: invalid value in motif call.')
			else:
				print(f'{line_num}: no value in motif call.')
		elif c == ':':
			# トラックループ
			pos += 1

			for i, ch in enumerate(score.ch):
				if ch_bits & (1 << i):
					ch.appendEvent(TYPE_LABEL, 0, 0, 0)
		elif c == '[':
			# リピート(開始)
			pos += 1

			for i, ch in enumerate(score.ch):
				if ch_bits & (1 << i):
					if ch.nest < 1:
						ch.assigned_label = ch.label
						ch.exist_skip = False
						ch.appendEvent(TYPE_LABEL, 0, ch.label, 0)
						ch.label += 1
						ch.nest += 1
					else:
						print(f'{line_num}: repeat nest over.')
		elif c == '|':
			# リピート(スキップ)
			pos += 1
			for i, ch in enumerate(score.ch):
				if ch_bits & (1 << i):
					ch.appendEvent(TYPE_SKIP, 0, ch.assigned_label, 0)
					ch.save_base_tone = ch.base_tone
					ch.save_def_len = ch.def_len
					ch.save_quant = ch.quant
					ch.exist_skip = True
		elif c == ']':
			# リピート(終了)
			pos += 1
			c = line[pos]
			if c.isdigit() or c == '$':
				pos, value = parseInt(line, pos)
				if value > 0 and value <= 127:
					for i, ch in enumerate(score.ch):
						if ch_bits & (1 << i):
							ch.nest -= 1
							ch.appendEvent(TYPE_REPEAT, value, ch.assigned_label, 0)
							if ch.exist_skip:
								ch.base_tone = ch.save_base_tone
								ch.def_len = ch.save_def_len
								ch.quant = ch.save_quant
				else:
					print(f'{line_num}: invalid value in repeat.')
			else:
				print(f'{line_num}: no value in repeat.')
		elif c == '(':
			# モチーフ範囲(開始)
			pos += 1

			for i, ch in enumerate(score.ch):
				if ch_bits & (1 << i):
					ch.appendEvent(TYPE_LABEL, 0, ch.label, 0)
					ch.ref_ev = ch.tail
					ch.label += 1
		elif c == ')':
			# モチーフ範囲(終了)
			pos += 1
			c = line[pos]
			if c.isdigit() or c == '$':
				pos, value = parseInt(line, pos)
				if value > 0:
					for i, ch in enumerate(score.ch):
						if ch_bits & (1 << i):
							if ch.ref_ev:
								ch.appendEvent(TYPE_RET, 0, ch.ref_ev.length, 0)
								ch.ref_ev.value = value
								ch.ref_ev = None
								if value > ch.max_ref_num:
									ch.max_ref_num = value
				else:
					print(f'{line_num}: invalid value in motif.')
			else:
				print(f'{line_num}: no value in motif.')
		elif c == '/':
			# トラック終端
			pos += 1

			for i, ch in enumerate(score.ch):
				if ch_bits & (1 << i):
					ch.appendEvent(TYPE_END, 0, 0, 0)
		else:
			# 未定義
			pos += 1
			print(f'{line_num}: syntax error. ({c})')

		c = line[pos]

def parse(score, ifp):
	line_num = 0
	for line in ifp:
		line_num += 1
		line += '\0\0'

		ch_bits = 0
		def_name = ''
		pos = 0
		c = line[pos]
		while c != '\0':
			if c.isspace():
				# チャンネル指定の区切り
				if ch_bits != 0:
					break
				pos += 1
			elif c == ';':
				# コメント
				pos = len(line)-1
			elif (index := CH_CHR.find(c)) >= 0:
				# チャンネル指定
				ch_bits |= (1 << index)
				pos += 1
			elif c == '#':
				# データ定義
				pos += 1
				i = 0
				while line[pos+i].isalnum():
					i += 1
				def_name = line[pos:pos+i]
				pos += i
				if line[pos].isspace():
					break
				print(f'{line_num}: unknown define name.')
			else:
				print(f'{line_num}: unknown channel. ({c})')
				pos += 1
			c = line[pos]

		if def_name:
			parseDefine(score, def_name, line, pos, line_num)
		else:
			parseScore(score, ch_bits, line, pos, line_num)

# 1. NA    LR NB NC SK ND NE RP NF
# 2. NA JS LR NB NC SK ND NE RP NF
# 3. NA JS LR NB NC LS ND NE RP NF
# pos_head <- NB
# LR.next <- LS.next (ND)
# pos_tail <- RP
# NE.next <- LS
# LS.next <- pos_head (NB)
# NC.next <- pos_tail (RP)
# 4. NA JS LR ND NE LS NB NC RP NF

def insertLoopEvent(score):
	for ch in score.ch:
		head_prv = [ None ] * ch.label
		head_cur = [ None ] * ch.label
		skip_prv = [ None ] * ch.label
		tail_prv = [ None ] * ch.label
		track_loop = False
		end_event = None
		prev = None
		ev = ch.head
		while ev:
			if ev.type == TYPE_LABEL and ev.length == 0:
				track_loop = True
			elif ev.type == TYPE_END:
				end_event = ev

			if ev.type == TYPE_LABEL:
				head_prv[ev.length] = prev
				head_cur[ev.length] = ev
			elif ev.type == TYPE_SKIP or ev.type == TYPE_SKIP_W:
				skip_prv[ev.length] = prev
			elif ev.type == TYPE_REPEAT or ev.type == TYPE_REPEAT_W or ev.type == TYPE_RET:
				tail_prv[ev.length] = prev
			prev = ev
			ev = ev.next

		if track_loop:
			if end_event:
				end_event.type = TYPE_JMP
			else:
				ch.appendEvent(TYPE_JMP, 0, 0, 0)
		else:
			if prev:
				if prev.type != TYPE_JMP and prev.type != TYPE_END:
					ch.appendEvent(TYPE_END, 0, 0, 0)

		# スキップありリピートを変換する
		for k in range(1,len(tail_prv)):
			if skip_prv[k]:
				ev = tail_prv[k].next
				if ev.type == TYPE_REPEAT or ev.type == TYPE_REPEAT_W:
					# ジャンプイベントを追加
					prev = head_prv[k]
					while prev.next != head_cur[k]:
						prev = prev.next
					ch.insertEvent(prev, TYPE_JMP, 0, ch.label, 0)
					# スキップイベントをラベルに変更
					ev = skip_prv[k].next
					ev.type = TYPE_LABEL
					ev.length = ch.label
					ch.label += 1
					# リピート先頭から元スキップまでと
					# 元スキップからリピート末尾までを入れ替える
					ev = prev.next.next
					repeat_head = ev.next
					ev.next = skip_prv[k].next.next
					ev = tail_prv[k]
					repeat_tail = ev.next
					ev.next = skip_prv[k].next
					skip_prv[k].next.next = repeat_head
					skip_prv[k].next = repeat_tail

def insertMotifLabelEvent(score):
	for ch in score.ch:
		tbl = [ None ] * (ch.max_ref_num + 1)

		# モチーフ範囲(開始)のラベルを抽出
		ev = ch.head
		while ev:
			if ev.type == TYPE_LABEL and ev.value > 0:
				tbl[ev.value] = call_t(ev)
			ev = ev.next

		# モチーフの呼び出し回数分、LABELを追加する
		ev = ch.head
		while ev:
			if ev.type == TYPE_CALL:
				ref_ev = tbl[ev.value].ev
				if ref_ev:
					if tbl[ev.value].count > 0:
						ev.length = ch.label
						ch.insertEvent(ref_ev, TYPE_LABEL, ev.value, ev.length, 0)
						ch.label += 1
					else:
						ev.length = ref_ev.length
					tbl[ev.value].count += 1
			ev = ev.next

def insertToneEvent(score):
	for ch in score.ch:
		low = 0
		high = 0
		last_low = 0
		last_high = 0
		sect = None
		prev = None
		ev = ch.head
		while ev:
			if ev.type == TYPE_NOTE:
				if low == 0:
					sect = prev
					low = ev.value
					high = ev.value
				else:
					if ev.value < low:
						if ev.value <= high - TONE_RANGE:
							ch.insertEvent(sect, TYPE_TONE_BASE, low, high, 0)
							sect = prev
							low = ev.value
							high = ev.value
						else:
							low = ev.value
					elif ev.value > high:
						if ev.value >= low + TONE_RANGE:
							ch.insertEvent(sect, TYPE_TONE_BASE, low, high, 0)
							sect = prev
							low = ev.value
							high = ev.value
						else:
							high = ev.value
			elif ev.type == TYPE_LABEL:
				if low != 0:
					ch.insertEvent(sect, TYPE_TONE_BASE, low, high, 0)
					sect = prev
					last_low = low
					last_high = high
					low = 0
					high = 0
			elif ev.type == TYPE_JMP or ev.type == TYPE_CALL:
				if low == 0 and last_low != 0:
					ch.insertEvent(prev, TYPE_TONE_BASE, last_low, last_high, 0)
					sect = ev
			prev = ev
			ev = ev.next
		if low > 0:
			ch.insertEvent(sect, TYPE_TONE_BASE, low, high, 0)

def mergeLengthTable(tbl1, tbl2):
	count = tbl1.commonCount(tbl2)
	if count == tbl1.count:
		# 共通する要素の数がtbl1の要素数と同じならばtbl2に統合
		tbl1.count = 0
		return 2
	elif count == tbl2.count:
		# 共通する要素の数がtbl2の要素数と同じならばtbl1に統合
		tbl2.count = 0
		return 1
	elif (tbl2.count + tbl1.count - count) <= LENTBL_SIZE:
		# tbl1とtbl2の統合後の要素数がLENTBL_SIZE内に収まるのであれば
		# tbl1に統合
		tbl1.merge(tbl2)
		tbl2.count = 0
		return 1

	return 0

def shareLengthTable(tbl1, tbl2):
	length = []
	quant = []
	count = 0
	mask1 = 0
	mask2 = 0
	i = 0
	while i < tbl1.count:
		j = 0
		while j < tbl2.count:
			if tbl1.length[i] == tbl2.length[j] and tbl1.quant[i] == tbl2.quant[j]:
				mask1 |= (1 << i)
				mask2 |= (1 << j)
				length.append(tbl1.length[i])
				quant.append(tbl1.quant[i])
				count += 1
			j += 1
		i += 1

	if count > 0:
		tbl1.share_count = count

		j = 0
		i = 0
		while i < tbl1.count:
			if (mask1 & (1 << i)) == 0:
				tbl1.length[j] = tbl1.length[i]
				tbl1.quant[j] = tbl1.quant[i]
				j += 1
			i += 1
		i = 0
		while i < count:
			tbl1.length[j] = length[i]
			tbl1.quant[j] = quant[i]
			j += 1
			i += 1

		j = tbl2.count - 1
		i = tbl2.count - 1
		while i >= 0:
			if (mask2 & (1 << i)) == 0:
				tbl2.length[j] = tbl2.length[i]
				tbl2.quant[j] = tbl2.quant[i]
				j -= 1
			i -= 1
		i = count - 1
		while i >= 0:
			tbl2.length[j] = length[i]
			tbl2.quant[j] = quant[i]
			j -= 1
			i -= 1

def insertLengthTableEvent(score):
	# 音長テーブルの作成
	for ch in score.ch:
		tbl1 = None
		sect = None
		prev = None
		ev = ch.head
		while ev:
			if ev.type == TYPE_NOTE or ev.type == TYPE_REST or ev.type == TYPE_EXTEND:
				if tbl1 == None:
					sect = prev
					tbl1 = lentbl_t()
					score.lentbl.append(tbl1)
					tbl1.append(ev.length, ev.quant)
				else:
					if not tbl1.append(ev.length, ev.quant):
						ch.insertEvent(sect, TYPE_LEN_TABLE, len(score.lentbl)-1, 0 ,0)
						sect = prev
						tbl1 = lentbl_t()
						score.lentbl.append(tbl1)
						tbl1.append(ev.length, ev.quant)
			elif ev.type == TYPE_LABEL:
				if tbl1:
					ch.insertEvent(sect, TYPE_LEN_TABLE, len(score.lentbl)-1, 0 ,0)
					sect = prev
					tbl1 = None
			elif ev.type == TYPE_JMP or ev.type == TYPE_CALL:
				if not tbl1 and sect:
					ch.insertEvent(prev, TYPE_LEN_TABLE, len(score.lentbl)-1, 0 ,0)
					sect = ev
			prev = ev
			ev = ev.next
		if tbl1:
			ch.insertEvent(sect, TYPE_LEN_TABLE, len(score.lentbl)-1, 0 ,0)

	# 音長テーブルの結合
	n = len(score.lentbl)
	for i in range(n):
		tbl1 = score.lentbl[i]
		if tbl1.count > 0:
			for j in range(n):
				tbl2 = score.lentbl[j]
				if tbl2 != tbl1 and tbl2.count > 0:
					r = mergeLengthTable(tbl1, tbl2)
					if r == 1:
						# tbl2を指しているイベントをtbl1を指すように変更
						for ch in score.ch:
							ev = ch.head
							while ev:
								if ev.type == TYPE_LEN_TABLE and ev.value == j:
									ev.value = i
								ev = ev.next
					elif r == 2:
						# tbl1を指しているイベントをtbl2を指すように変更
						for ch in score.ch:
							ev = ch.head
							while ev:
								if ev.type == TYPE_LEN_TABLE and ev.value == i:
									ev.value = j
								ev = ev.next

	# 隣り合う音長テーブルに共通する音長を抽出
	tbl1 = None
	for tbl2 in score.lentbl:
		if tbl2.count > 0:
			if tbl1:
				shareLengthTable(tbl1, tbl2)
				tbl1 = None
			else:
				tbl1 = tbl2

	# 音長テーブルのベース位置を設定
	pos = len(score.rhythm)
	for tbl1 in score.lentbl:
		tbl1.base = pos
		pos += tbl1.count - tbl1.share_count

def encodeEnvelopeTable(score):
	# vvvv0000
	# vvvv0010 ccccdddd
	# vvvv0001 ccccdddd dddddddd
	# envelope loop instruction when the 1st byte is $d0 or more

	for env in score.env:
		count = len(env.vol)
		if count > 0:
			env.data.append(env.vol[0] & 255)
			loop = -1
			i = 1
			while i < count:
				if i == env.loop:
					loop = len(env.data)
				n = ((env.vol[i] + 4) & 15) << 4
				if env.dt[i] == 0 and env.tc[i] == 0:
					env.data.append(n & 255)
				else:
					if env.dt[i] >= -8 and env.dt[i] < 8:
						n |= 1
						env.data.append(n & 255)
						n = ((env.tc[i] & 15) << 4) | (env.dt[i] & 15)
						env.data.append(n & 255)
					else:
						n |= 2
						env.data.append(n & 255)
						n = ((env.tc[i] & 15) << 12) | (env.dt[i] & 4095)
						env.data.append((n >> 8) & 255)
						env.data.append( n       & 255)
				i += 1
			if loop >= 0:
				n = loop - (len(env.data) + 1)
				if n < -48:
					print(f'envelop loop distance error. (EP {score.env.find(env)})')
			else:
				n = -1
			env.data.append(n & 255)

def optimizeEvent(score):
	for ch in score.ch:
		while True:
			tbl = []
			for i in range(ch.label):
				tbl.append(opttbl_t())

			low = 0
			high = 0
			uninit_tone = True
			ltbl = None
			cont_len = True
			label1 = None
			label2 = None
			prev = None #-2
			ev = ch.head
			while ev:
				if ev.type == TYPE_LABEL:
					tbl[ev.length].link1 = label1
					tbl[ev.length].link2 = label2
					label1 = tbl[ev.length]
					label2 = tbl[ev.length]
					label1.near.ev = prev
					label1.near.low = low
					label1.near.high = high
					label1.near.uninit_tone = uninit_tone
					label1.near.ltbl = ltbl
					label1.near.cont_len = cont_len
				elif ev.type == TYPE_REPEAT or ev.type == TYPE_REPEAT_W or ev.type == TYPE_CALL:
					tbl[ev.length].far.ev = prev
					tbl[ev.length].far.low = low
					tbl[ev.length].far.high = high
					tbl[ev.length].far.uninit_tone = uninit_tone
					tbl[ev.length].far.ltbl = ltbl
					tbl[ev.length].far.cont_len = cont_len
				elif ev.type == TYPE_JMP:
					tbl[ev.length].far.ev = prev
					tbl[ev.length].far.low = low
					tbl[ev.length].far.high = high
					tbl[ev.length].far.uninit_tone = uninit_tone
					tbl[ev.length].far.ltbl = ltbl
					tbl[ev.length].far.cont_len = cont_len
					low = 0
					high = 0
					ltbl = None
					uninit_tone = False
					cont_len = False
					label1 = None
					label2 = None
					prev = None
				elif ev.type == TYPE_TONE_BASE:
					prev = ev
					low = ev.value
					high = ev.length
					uninit_tone = False
				elif ev.type == TYPE_LEN_TABLE:
					ltbl = score.lentbl[ev.value]
					cont_len = True
					label2 = None
				elif ev.type == TYPE_NOTE:
					label1 = None
					label2 = None
				ev = ev.next

			ltbl = None
			cont_len = True
			ev = ch.head
			while ev:
				if ev.type == TYPE_LABEL:
					label1 = tbl[ev.length]
				elif ev.type == TYPE_JMP:
					ltbl = None
					cont_len = False
					label1 = None
				elif ev.type == TYPE_TONE_BASE:
					if label1:
						low = ev.value
						high = ev.length
						label2 = label1
						while label2:
							if label2.near.ev and label2.link1 == None:
								if label2.near.low < low:
									low = label2.near.low
								if label2.near.high > high:
									high = label2.near.high
							if label2.far.ev:
								if label2.far.low < low:
									low = label2.far.low
								if label2.far.high > high:
									high = label2.far.high
							label2 = label2.link1

						if high - low + 1 <= TONE_RANGE:
							do_break = False
							ev.quant = 1
							if ev.value != low:
								do_break = True
							if ev.length != high:
								do_break = True
							ev.value = low
							ev.length = high
							label2 = label1
							while label2:
								if label2.near.uninit_tone:
									ev.quant = 0
								if label2.near.ev and label2.link1 == None:
									if label2.near.ev.value != low:
										do_break = True
									if label2.near.ev.length != high:
										do_break = True
									label2.near.ev.value = low
									label2.near.ev.length = high

								if label2.far.uninit_tone:
									ev.quant = 0
								if label2.far.ev:
									if label2.far.ev.value != low:
										do_break = True
									if label2.far.ev.length != high:
										do_break = True
									label2.far.ev.value = low
									label2.far.ev.length = high

								label2 = label2.link1
							if do_break:
								break
						else:
							ev.quant = 0
					low = ev.value
					high = ev.length
				elif ev.type == TYPE_LEN_TABLE:
					ltbl2 = score.lentbl[ev.value]
					if label1 and ev.quant == 0:
						ev.quant = 1
						if cont_len and ltbl2 != ltbl:
							ev.quant = 0
						label2 = label1
						while label2:
							if label2.near.cont_len and ltbl2 != label2.near.ltbl:
								ev.quant = 0
							if label2.far.cont_len and ltbl2 != label2.far.ltbl:
								ev.quant = 0
							label2 = label2.link2
					ltbl = ltbl2
					cont_len = True
				elif ev.type == TYPE_NOTE:
					label1 = None
				ev = ev.next
			if ev == None:
				break

def assignLabelOffset(score):
	for ch in score.ch:
		locate = [ 0 ] * ch.label
		for i in range(3):
			count = 0
			ev = ch.head
			while ev:
				if ev.type == TYPE_NOTE:
					count += 1
				elif ev.type == TYPE_REST:
					count += 1
				elif ev.type == TYPE_EXTEND:
					count += 1
				elif ev.type == TYPE_REG1:
					count += 2
				elif ev.type == TYPE_REG2:
					count += 2
				elif ev.type == TYPE_TONE_BASE and ev.quant == 0:
					count += 2
				elif ev.type == TYPE_LEN_TABLE and ev.quant == 0:
					count += 2
				elif ev.type == TYPE_ENV_BASE:
					count += 2
				elif ev.type == TYPE_TEMPO and ev.quant == 0:
					count += 2
				elif ev.type == TYPE_PERCUSS:
					count += 2
				elif ev.type == TYPE_VOLUME:
					count += 1
				elif ev.type == TYPE_LOOP_CNT:
					count += 1
				elif ev.type == TYPE_REPEAT:
					n = locate[ev.length] - (count + 1)
					if ev.value == 2 and n <= -3 and n >= -4:
						count += 1
						ev.quant = 0
					elif ev.value == 3 and n <= -2 and n >= -4:
						count += 1
						ev.quant = 0
					elif ev.value == 4 and n == -2:
						count += 1
						ev.quant = 0
					elif ev.value <= 9 and n > -32:
						count += 2
						ev.quant = 1
						n = locate[ev.length] - count
					else:
						count += 3
						ev.quant = 2
						n = locate[ev.length] - count
					if i == 2:
						if ev.value <= 127 and n > -256:
							ev.length = (locate[ev.length] - count) & 255
						else:
							if ev.value > 127:
								print(f'repeat count error.')
							if n <= -256:
								print(f'repeat distance error.')
				elif ev.type == TYPE_REPEAT_W:
					n = locate[ev.length] - (count + 1)
					if ev.value <= 9 and n >= -32:
						count += 2
						ev.quant = 1
						n = locate[ev.length] - count
					else:
						count += 3
						ev.quant = 2
						n = locate[ev.length] - count
					if i == 2:
						if ev.value <= 127 and n > -256:
							ev.length = (locate[ev.length] - count) & 0xff
						else:
							if ev.value > 127:
								print(f'repeat count error.')
							if n <= -256:
								print(f'repeat distance error.')
				elif ev.type == TYPE_JMP or ev.type == TYPE_CALL:
					if i == 0:
						count += 3
					elif i == 1:
						n = locate[ev.length] - (count + 2)
						if n > -64 and n < 64:
							count += 2
						else:
							count += 3
					else:
						n = locate[ev.length] - (count + 2)
						if n > -64 and n < 64:
							ev.quant = 0
							count += 2
						else:
							ev.quant = 1
							count += 3
						ev.length = (locate[ev.length] - count) & 0xffff
				elif ev.type == TYPE_RET:
					count += 1
				elif ev.type == TYPE_END:
					count += 1
				elif ev.type == TYPE_LABEL:
					if i < 2:
						locate[ev.length] = count
				ev = ev.next

def assignEnvTableBase(score):
	count = 2
	for env in score.env:
		env.base = count
		count += len(env.data)

	for rhythm in score.rhythm:
		n = rhythm.data & 15
		if n > 0:
			n = score.env[n-1].base
			if n <= 15:
				rhythm.data &= 0xfff0
				rhythm.data |= n
			else:
				print(f'rhythm envelop error. ({ryhthm.name})')

def writeCmd(score, cfp, prefix):
	valid = (score.tempo_1st & 255) << 8
	for i, ch in enumerate(score.ch):
		if ch.head:
			valid |= 1 << i

	cfp.write(f'{prefix}_PlayData:\n')
	cfp.write(f'\t.dw\t${valid:04x}, {prefix}_LenTable-4, {prefix}_EnvTable\n')
	first = True
	for i, ch in enumerate(score.ch):
		if ch.head:
			if first:
				cfp.write(f'\t.dw\t{prefix}_Seq{chr(i+65)}')
				first = False
			else:
				cfp.write(f', {prefix}_Seq{chr(i+65)}')
	cfp.write(f'\n')

def writeData(score, dfp, prefix):
	# 音長テーブル
	dfp.write(f'{prefix}_LenTable:\n')
	count = 0
	for rhythm in score.rhythm:
		if (count % 4) == 0:
			dfp.write(f'\t.dw\t')
		else:
			dfp.write(f', ')
		dfp.write(f'${rhythm.data:04x}')
		if (count % 4) == 3:
			dfp.write(f'\n')
		count += 1
	if (count % 4) != 0:
		dfp.write(f'\n')

	tick_min = (score.tempo_max+75-1) // 75
	count = 0
	for tbl in score.lentbl:
		for i in range(tbl.count - tbl.share_count):
			if (count % 4) == 0:
				dfp.write(f'\t.db\t')
			else:
				dfp.write(f', ')
			if tbl.quant[i] == 8 or tbl.length[i] <= tick_min:
				n = 0
			else:
				n = tbl.length[i] * (8 - tbl.quant[i]) // 8
				if n < tick_min:
					if tbl.length[i] - tick_min < tick_min:
						n = 0
					else:
						n = tick_min
			dfp.write(f'{tbl.length[i]-n:3}, {n:3}')
			if (count % 4) == 3:
				dfp.write(f'\n')
			count += 1
	if (count % 4) != 0:
		dfp.write(f'\n')

	# エンベロープテーブル
	dfp.write(f'{prefix}_EnvTable:\n')
	dfp.write(f'\t.db\t$00, $ff')
	count = 2
	for env in score.env:
		for data in env.data:
			if (count % 8) == 0:
				dfp.write(f'\t.db\t')
			else:
				dfp.write(f', ')
			dfp.write(f'${data:02x}')
			if (count % 8) == 7:
				dfp.write(f'\n')
			count += 1
	if (count % 8) != 0:
		dfp.write(f'\n')

	# シーケンスデータ
	for i, ch in enumerate(score.ch):
		if ch.head:
			#print(f'{prefix}_Seq{chr(i+65)}:')
			dfp.write(f'{prefix}_Seq{chr(i+65)}:\n')
			buf = []
			count = 0
			tone_base = 0
			tbl = None
			ev = ch.head
			while ev:
				if ev.type == TYPE_NOTE:
					i = 0
					if tbl:
						while i < len(tbl.length):
							if tbl.length[i] == ev.length and tbl.quant[i] == ev.quant:
								break
							i += 1
						if i >= len(tbl.length):
							i = 0
					else:
						i = 0
					buf = [ ((i+2) << 4) | (ev.value - tone_base + 2) ]
				elif ev.type == TYPE_REST:
					i = 0
					if tbl:
						while i < len(tbl.length):
							if tbl.length[i] == ev.length and tbl.quant[i] == ev.quant:
								break
							i += 1
						if i >= len(tbl.length):
							i = 0
					else:
						i = 0
					buf = [ ((i+2) << 4) | 0x01 ]
				elif ev.type == TYPE_EXTEND:
					i = 0
					if tbl:
						while i < len(tbl.length):
							if tbl.length[i] == ev.length and tbl.quant[i] == ev.quant:
								break
							i += 1
						if i >= len(tbl.length):
							i = 0
					else:
						i = 0
					buf = [ ((i+2) << 4) ]
				elif ev.type == TYPE_REG1:
					buf = [ 0x10, ev.value ]
				elif ev.type == TYPE_REG2:
					buf = [ 0x11, ev.value ]
				elif ev.type == TYPE_TONE_BASE:
					tone_base = ev.value
					if ev.quant == 0:
						buf = [ 0x12, tone_base-2*12 ]
					else:
						buf = []
				elif ev.type == TYPE_LEN_TABLE:
					tbl = score.lentbl[ev.value]
					if ev.quant == 0:
						buf = [ 0x13, tbl.base ]
					else:
						buf = []
				elif ev.type == TYPE_ENV_BASE:
					if ev.value > 0:
						n = score.env[ev.value-1].base
					else:
						n = 0
					buf = [ 0x14, n ]
				elif ev.type == TYPE_TEMPO:
					if ev.quant == 0:
						buf = [ 0x15, ev.value ]
				elif ev.type == TYPE_PERCUSS:
					buf = [ 0x16, ev.value ]
				elif ev.type == TYPE_VOLUME:
					if ev.value > 0:
						buf = [ 0x16 + ev.value*2 ] # +1:0x18 +2:0x1a
					else:
						buf = [ 0x17 + (0-ev.value)*2 ] # -1:0x19 -2:9x1b
				elif ev.type == TYPE_LOOP_CNT:
					buf = [ 0x17 ]
				elif ev.type == TYPE_CALL:
					if ev.quant == 0:
						buf = [ 0x02, ev.length&0x7f ]
					else:
						buf = [ 0x02, (ev.length&0x7f)|0x80, (ev.length>>7)&0xff ]
				elif ev.type == TYPE_RET:
					buf = [ 0x01 ]
				elif ev.type == TYPE_REPEAT:
					if ev.quant == 0:
						if ev.length == 0xfe:
							buf = [ 0x05 - (ev.value&1) ]
						elif ev.length == 0xfd:
							buf = [ 0x06 + (ev.value&1) ]
						elif ev.length == 0xfc:
							buf = [ 0x08 + (ev.value&1) ]
						else:
							buf = []
					elif ev.quant == 1:
						buf = [ 0x0a, (((ev.value-2)<<5)|(ev.length&0x1f))&0xff ]
					else:
						buf = [ 0x0b, ev.value & 0xff, ev.length & 0xff ]
				elif ev.type == TYPE_REPEAT_W:
					if ev.quant == 1:
						buf = [ 0x0c, (((ev.value-2)<<5)|(ev.length&0x1f))&0xff ]
					else:
						buf = [ 0x0d, ev.value & 0xff, ev.length & 0xff ]
				elif ev.type == TYPE_JMP:
					if ev.quant == 0:
						buf = [ 0x03, ev.length&0x7f ]
					else:
						buf = [ 0x03, (ev.length&0x7f)|0x80, (ev.length>>7)&0xff ]
				elif ev.type == TYPE_END:
					buf = [ 0x00 ]
				else:
					buf = []

				#s = f'{ev.type} [ '
				#for b in buf:
				#	s += f'{b:02x} '
				#s += f']'
				#print(s)

				for n in buf:
					if (count % 8) == 0:
						dfp.write(f'\t.db\t')
					else:
						dfp.write(f', ')
					dfp.write(f'${n:02x}')
					if (count % 8) == 7:
						dfp.write(f'\n')
					count += 1
				ev = ev.next
			if (count % 8) != 0:
				dfp.write(f'\n')

def dumpEvent(ev):
	note = [ 'c ', 'c+', 'd ', 'd+', 'e ', 'f ', 'f+', 'g ', 'g+', 'a ', 'a+', 'b ' ]
	while ev:
		if ev.type == TYPE_NOTE:
			tone = ev.value
			print(f'(o{tone//12}{note[tone%12]}) {ev.length:3}')
		elif ev.type == TYPE_REST:
			print(f'(r)    {ev.length:3}')
		elif ev.type == TYPE_EXTEND:
			print(f'(x)    {ev.length:3}')
		elif ev.type == TYPE_REG1:
			print(f'(reg1) ')
		elif ev.type == TYPE_REG2:
			print(f'(reg2) ')
		else:
			print(f'({ev.type})')
		ev = ev.next

def dump(score):
	for i, ch in enumerate(score.ch):
		print(f'==== ch{chr(65+i)} ====')
		dumpEvent(ch.head)

class stat_len_t:
	def __init__(self, length = 0, quant = 0):
		self.count = 1
		self.length = length
		self.quant = quant

def statLength(stat_len, ev):
	e = next((e for e in stat_len if e.length == ev.length and e.quant == ev.quant), None)
	if e:
		e.count += 1
	else:
		stat_len.append(stat_len_t(ev.length, ev.quant))

def analyzeEvent(ev):
	count_total = 0
	count_note = 0
	count_extend = 0
	count_tone_base = 0
	count_len_table = 0
	stat_len = []
	while ev:
		count_total += 1
		if ev.type == TYPE_NOTE:
			count_note += 1
			statLength(stat_len, ev)
		elif ev.type == TYPE_REST:
			count_note += 1
			statLength(stat_len, ev)
		elif ev.type == TYPE_EXTEND:
			count_extend += 1
			statLength(stat_len, ev)
		elif ev.type == TYPE_TONE_BASE and ev.quant == 0:
			count_tone_base += 1
		elif ev.type == TYPE_LEN_TABLE and ev.quant == 0:
			count_len_table += 1
		elif ev.type == TYPE_LABEL or ev.type == TYPE_SKIP or ev.type == TYPE_SKIP_W:
			count_total -= 1
		ev = ev.next
	print(f'total:     {count_total:5}')
	print(f'note:      {count_note:5} ({count_note / count_total * 100.0:6.2f})')
	print(f'extend:    {count_extend:5} ({count_extend / count_total * 100.0:6.2f})')
	print(f'tone_base: {count_tone_base:5} ({count_tone_base / count_total * 100.0:6.2f})')
	print(f'len_table: {count_len_table:5} ({count_len_table / count_total * 100.0:6.2f})')

	count_len_total = 0
	for s in stat_len:
		count_len_total += s.count
	for s in stat_len:
		print(f'[{s.length:3} {s.quant:1}] {s.count:5} ({s.count / count_len_total * 100.0:6.2f})')

def analyze(score):
	for i, ch in enumerate(score.ch):
		print(f'==== ch{chr(65+i)} ====')
		analyzeEvent(ch.head)

def main():
	# 引数を解析
	params = parseArg()

	mmlName = params.mmlfile
	cmdName = params.cmdfile
	datName = params.datfile
	prefix  = params.prefix

	score = score_t()
	with open(mmlName, 'r') as ifp:
		with open(cmdName, 'w') as cfp:
			with open(datName, 'w') as dfp:
				parse(score, ifp)
				insertLoopEvent(score)
				insertMotifLabelEvent(score)
				insertToneEvent(score)
				insertLengthTableEvent(score)
				encodeEnvelopeTable(score)
				optimizeEvent(score)
				assignLabelOffset(score)
				assignEnvTableBase(score)
				writeCmd(score, cfp, prefix)
				writeData(score, dfp, prefix)
				#dump(score)
				#analyze(score)

main()
